from sqlalchemy import DateTime
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression


class UtcNow(expression.FunctionElement):
    type = DateTime()


@compiles(UtcNow, 'postgresql')
def pg_utc_now(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"



from fastapi import Response

__all__ = ('nginx_cache',)


def nginx_cache(timeout: int):
    def nginx_dependence(response: Response):
        response.headers['X-Accel-Expires'] = str(timeout)
        return response
    return nginx_dependence



from fastapi_toolkit.crud.base import CRUDBase
from fastapi_toolkit.db import create_session
from tests.models import Book


class TestCRUDGet:
    async def _create_book(self, **kwargs):
        async with create_session() as session:
            session.add(Book(**kwargs))
            await session.commit()

    async def test_one_ok(self, db):
        test_name = 'test_name'

        await self._create_book(name=test_name)

        book = await CRUDBase(Book).get(Book.name == test_name)
        assert book.name == test_name

    async def test_no_one_ok(self, db):
        await self._create_book(name='test_name')

        book = await CRUDBase(Book).get(Book.name == 'wrong_name')
        assert book is None



import typer
import uvicorn

from fastapi_toolkit.conf import settings


def runserver(
        host: str = typer.Option('0.0.0.0'),
        port: int = typer.Option(4001),
        debug: bool = typer.Option(False),
        reload: bool = typer.Option(False),
        workers: int = typer.Option(1)
):
    uvicorn.run(
        settings.application, host=host, port=port,
        debug=debug, reload=reload, workers=workers
    )



from pydantic import AnyUrl

from fastapi_toolkit.conf.global_settings import GlobalSettings


class Settings(GlobalSettings):
    project: str = 'test_project'
    # TODO: fix database settings
    database_dsn: AnyUrl = 'postgresql+asyncpg://postgres:postgres@127.0.0.1:5432/test_fastapi_toolkit'  # noqa



from datetime import datetime

from fastapi import (
    APIRouter,
    FastAPI,
)
from fastapi.middleware.httpsredirect import HTTPSRedirectMiddleware
from fastapi.openapi.utils import get_openapi
from fastapi.responses import ORJSONResponse

from fastapi_toolkit.conf import settings
from fastapi_toolkit.db import init_db

__all__ = (
    'app',
)

from fastapi_toolkit.schemas.application_info import ApplicationInfo


def custom_openapi(app):
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title=settings.project,
        version=settings.version or '0.0.1',
        routes=app.routes,
    )
    openapi_schema['x-tagGroups'] = settings.tag_groups
    app.openapi_schema = openapi_schema
    return app.openapi_schema


debug_router = APIRouter()


@debug_router.get('/', response_model=ApplicationInfo)
def application_info():
    """Returns application info"""

    return {
        'project': settings.project,
        'version': settings.version,
        'git': {
            'hash': settings.git_hash,
            'branch': settings.git_branch
        },
        'datetime': datetime.utcnow().isoformat()
    }


app = FastAPI(
    title=settings.project,
    default_response_class=ORJSONResponse
)
if settings.use_https:
    app.add_middleware(HTTPSRedirectMiddleware)

app.openapi = lambda: custom_openapi(app)
app.include_router(debug_router, prefix='')


@app.on_event('startup')
async def startup():
    init_db()


@app.on_event('shutdown')
async def shutdown():
    pass



from typing import (
    List,
    Union,
)

from pydantic import (
    BaseModel,
    Field,
    create_model,
)


class BaseMeta(BaseModel):
    count: int


class PageNumberPagination(BaseMeta):
    page: int = Field(None, gt=0)
    page_size: int = Field(None, gt=0)


class LimitOffsetPagination(BaseMeta):
    limit: int = Field(None, gt=0)
    offset: int = Field(None, ge=0)


class Meta(BaseModel):
    __root__: Union[LimitOffsetPagination, PageNumberPagination]


def pagination_model(model, prefix=''):
    return create_model(
        f'{prefix}Pagination{model.__name__}',
        meta=(Meta, ...),
        result=(List[model], ...)
    )
